@@ -1,4 +1,4 @@
-/* Copyright (C) 2004,2005 Andi Kleen, SuSE Labs.
+/* Copyright (C) 2004,2005,2006 Andi Kleen, SuSE Labs.
    Decode IA32/x86-64 machine check events in /dev/mcelog. 
 
    mcelog is free software; you can redistribute it and/or
@@ -30,6 +30,7 @@
 #include "mcelog.h"
 #include "k8.h"
 #include "p4.h"
+#include "dmi.h"
 
 enum {
 	CPU_GENERIC,
@@ -40,6 +41,9 @@ enum {
 char *logfn = "/dev/mcelog";
 
 int use_syslog;
+int do_dmi;
+int ignore_nodev;
+int filter_bogus;
 
 void Wprintf(char *fmt, ...)
 {
@@ -67,8 +71,46 @@ char *bankname(unsigned bank)
 	}
 } 
 
+void resolveaddr(unsigned long addr)
+{
+	if (addr && do_dmi)
+		dmi_decodeaddr(addr);
+	/* Should check for PCI resources here too */
+}
+
+int mce_filter_k8(struct mce *m)
+{	
+	/* Filter out GART errors */
+	if (m->bank == 4) { 
+		unsigned short exterrcode = (m->status >> 16) & 0x0f;
+		if (exterrcode == 5 && (m->status & (1ULL<<61)))
+			return 0;
+	} 
+	return 1;
+}
+
+int mce_filter(struct mce *m)
+{
+	if (!filter_bogus) 
+		return 1;
+	/* Filter out known broken MCEs */
+	switch (cpu) {
+	case CPU_K8:
+		return mce_filter_k8(m);
+		/* add more buggy CPUs here */
+	case CPU_P4:
+		/* No bugs known */
+		return 1;
+	default:
+	case CPU_GENERIC:
+		return 1;
+	}	
+}
+
 void dump_mce(struct mce *m) 
 {
+	Wprintf("HARDWARE ERROR. This is *NOT* a software problem!\n");
+	Wprintf("Please contact your hardware vendor\n");
 	/* should not happen */
 	if (!m->finished)
 		Wprintf("not finished?\n");
@@ -78,8 +120,6 @@ void dump_mce(struct mce *m)
 			m->tsc, 
 			(m->mcgstatus & MCI_STATUS_UC) ? 
 			"(upper bound, found by polled driver)" : "");
-	else
-		Wprintf("from boot or resume\n");
 	if (m->rip) 
 		Wprintf("RIP%s %02x:%Lx ", 
 		       !(m->mcgstatus & MCG_STATUS_EIPV) ? " !INEXACT!" : "",
@@ -103,6 +143,7 @@ void dump_mce(struct mce *m)
 	} 
 	/* decode all status bits here */
 	Wprintf("STATUS %Lx MCGSTATUS %Lx\n", m->status, m->mcgstatus);
+	resolveaddr(m->addr);
 }
 
 void check_cpu(void)
@@ -169,12 +210,17 @@ void decodefatal(FILE *inf)
 	char symbol[100];
 	int data = 0;
 
+	if (do_dmi)
+		Wprintf(
+ "WARNING: with --dmi mcelog --ascii must run on the same machine with the\n"
+ "     same BIOS/memory configuration as where the machine check occurred.\n");
+
 	k = 0;
 	memset(&m, 0, sizeof(struct mce));
 	symbol[0] = '\0';
 	while (getdelim(&line, &linelen, '\n', inf) > 0) { 
 		int n = 0;
-		char *s = skipgunk(line);
+		char *s = skipgunk(line), *p;
 
 		if (!strncmp(s, "CPU", 3)) { 
 			unsigned cpu = 0, bank = 0;
@@ -185,12 +231,30 @@ void decodefatal(FILE *inf)
 			       &bank,
 			       &m.status
 			       );
-			m.cpu = cpu;
-			m.bank = bank;
-			if (n != 4) 
-				missing++; 
-		} 
+			if (n == 1) {
+				n = sscanf(s, "CPU %u %u", &cpu, &bank);
+				m.cpu = cpu;
+				m.bank = bank;
+				if (n != 2) 
+					missing++;
+				
+				if ((p = strstr(s, "TSC")) != NULL) {
+					if (sscanf(p,"TSC %Lx", &m.tsc)!=1)
+						missing++;
+				}
+			} else { 
+				m.cpu = cpu;
+				m.bank = bank;
+				if (n != 4) 
+					missing++; 
 
+			}
+		} 
+		else if (!strncmp(s, "STATUS", 6)) {
+			if (sscanf(s,"STATUS %Lx MCGSTATUS %Lx", 
+				   &m.status,  &m.mcgstatus) != 2)
+				missing++;			
+		} 
 		else if (!strncmp(s, "RIP", 3)) { 
 			unsigned cs = 0; 
 
@@ -214,7 +278,7 @@ void decodefatal(FILE *inf)
 		}
 		else if (!strncmp(s, "ADDR",4)) { 
 			if ((n = sscanf(s, "ADDR %Lx", &m.addr)) != 1) 
-				missing++; 
+				missing++;
 		}
 		else if (!strncmp(s, "MISC",4)) { 
 			if ((n = sscanf(s, "MISC %Lx", &m.misc)) != 1) 
@@ -239,54 +303,68 @@ void usage(void)
 {
 	fprintf(stderr, 
 		"Usage:\n"
-		"  mcelog [--k8|--p4|--generic] [--syslog] [mcelogdevice]\n"
-		"  mcelog [--k8|--p4|--generic] --ascii\n"
-		"Decode machine check error records\n");	
+		"  mcelog [--k8|--p4|--generic] [--ignorenodev] [--dmi] [--syslog] [--filter] [mcelogdevice]\n"
+		"Decode machine check error records from kernel\n"
+		"  mcelog [--k8|--p4|--generic] [--dmi] --ascii < log\n"
+		"Decode machine check ASCII output from kernel logs\n");
 	exit(1);
 }
 
+int modifier(char *s)
+{
+	if (!strcmp(s, "--k8")) {
+		cpu = CPU_K8;
+	} else if (!strcmp(s, "--p4")) {
+		cpu = CPU_P4;
+	} else if (!strcmp(s, "--generic")) { 
+		cpu = CPU_GENERIC;
+	} else if (!strcmp(s, "--ignorenodev")) { 
+		ignore_nodev = 1;
+	} else if (!strcmp(s,"--filter")) { 
+		filter_bogus = 1;			
+	} else if (!strcmp(s, "--dmi")) { 
+		do_dmi = 1;
+		opendmi();
+	} else if (!strcmp(s, "--syslog")) { 
+		openlog("mcelog", 0, LOG_DAEMON);
+		use_syslog = 1;
+	} else
+		return 0;
+	return 1;
+} 
+
 int main(int ac, char **av) 
 { 
 	int recordlen = 0;
 	int loglen = 0;
 
-	if (av[1] && !strcmp(av[1],"--syslog")) { 
-		av++;
-		openlog("mcelog", 0, LOG_DAEMON);
-		use_syslog = 1;
-	}
-
 	check_cpu();
-	if (av[1]) { 
-		if (!strcmp(av[1], "--k8")) {
-			cpu = CPU_K8;
-			av++;
-		} else if (!strcmp(av[1], "--p4")) {
-			cpu = CPU_P4;
-			av++;
-		} else if (!strcmp(av[1], "--generic")) { 
-			cpu = CPU_GENERIC;
-			av++;
-		}
-	}		
-	
-	if (av[1] && !strcmp(av[1], "--ascii")) { 
-		av++;
-		if (av[1]) 
-			usage();
-		decodefatal(stdin); 
-		exit(0);
-	} 
 
-	if (av[1]) {
-		logfn = av[1];
-		av++;
-	}
-	if (av[1])
+	while (*++av) { 
+		if (modifier(*av)) {
+			/* ok */
+		} else if (!strcmp(*av, "--ascii")) { 
+			while (*++av) {
+				if (!modifier(*av))
+					usage();
+			}
+			decodefatal(stdin); 
+			exit(0);
+		} else if (!strncmp(*av, "--", 2)) { 
+			if (av[0][2] == '\0') 
+				break;
+			usage();
+		}
+	} 
+	if (*av)
+		logfn = *av++;
+	if (*av)
 		usage();
 		
 	int fd = open(logfn, O_RDONLY); 
 	if (fd < 0) {
+		if (ignore_nodev) 
+			exit(0);
 		fprintf(stderr, "Cannot open %s\n", logfn); 
 		exit(1);
 	}
@@ -311,6 +389,8 @@ int main(int ac, char **av)
 	int i; 
 	for (i = 0; i < len / recordlen; i++) { 
 		struct mce *mce = (struct mce *)(buf + i*recordlen);
+		if (!mce_filter(mce)) 
+			continue;
 		printf("MCE %d\n", i); 
 		dump_mce(mce); 
 	}
